char IDENT[]=
	"@(#) (C) Copyright 1994 by Christian Finger (finger@porsche.de)";
static char copyright[] = "@(#)Copyright (C) 2003-2018 by Andreas Ley (andreas.ley@kit.edu)";
static char sccsid[] = "@(#)lockcmd 1.1 - lock file while executing command";

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/wait.h>

#define SHELL "/bin/sh"

void usage(line)
char	*line;
{
	fprintf(stderr,"%s",line);
	fprintf(stderr,"Usage: lockcmd [ options ] lockfile \"command [parameter ...]\"\n");
	fprintf(stderr,"Or:    lockcmd [ options ] lockfile command [parameter ...]\n");
	fprintf(stderr,"  options are:\n");
	fprintf(stderr,"   -e     exclusive lock (default)\n");
	fprintf(stderr,"   -s     shared lock\n");
	fprintf(stderr,"   -w     wait if file is already locked (default)\n");
	fprintf(stderr,"   -n     no wait if file is already locked\n");
	fprintf(stderr,"   -h     display this help\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"lockcmd locks <lockfile> while <command> is executed.\n");
	fprintf(stderr,"\n");
	exit(1);
}

int main(argc,argv)
int		argc;
char	**argv;
{

	int				mode, lockcmd;
	int				stat_val;
	int				lf;				/* lock file */
	struct flock	lcs;			/* lock control structure */

	extern int optind;

	lcs.l_whence=SEEK_SET;
	lcs.l_start=0;
	lcs.l_len=0;

	/* set defaults */
	lcs.l_type=F_WRLCK;
	mode=O_WRONLY|O_CREAT;
	lockcmd=F_SETLKW;

	{
		int c;
		extern char *optarg;

		while ((c=getopt(argc, argv,"eswnh")) != EOF)
			switch(c) {
				case 'e':	/* exclusive lock */
							mode=O_WRONLY|O_CREAT;
							lcs.l_type=F_WRLCK;
							break;
				case 's':	/* shared lock */
							mode=O_RDONLY|O_CREAT;
							lcs.l_type=F_RDLCK;
							break;
				case 'w':	/* wait until lock is free */
							lockcmd=F_SETLKW;
							break;
				case 'n':	/* nowait */
							lockcmd=F_SETLK;
							break;
				case 'h':	usage("");
				case '?':	usage("Error: unknown option\n");
			}
	}

	if (argc-optind<2)
		usage("Error: Parameter count mismatch\n");
	
	if ((lf=open(argv[optind],mode,0600))<0) {
		fprintf(stderr,"Error opening lock file\n");
		perror("lockcommand/open");
		exit(1);
	}

	if (fcntl(lf,lockcmd,&lcs)==-1) {
		if ((errno == EACCES) || (errno == EAGAIN)) {
			/*file already locked by another process and we do not wait for it*/
			exit(255);
		}
		else {
			fprintf(stderr,"Error locking file\n");
			perror("lockcommand/fcntl");
			exit(1);
		}
	}

	switch (vfork()) {
		case -1	:	/* can't fork */
					perror("lockcommand/fork");
					exit(1);
					break;
		case  0	:	if (argc-optind>2) {
						char	**args;
						int		i;
						/* Can't guarantee argv[argc] is (char *)0 ... */
						args = malloc(sizeof(char *)*(argc-optind));
						for (i=0;i<argc-optind-1;i++)
							args[i] = argv[optind+i+1];
						args[argc-optind-1] = (char *)0;
						if (strrchr(args[0],'/'))
							args[0] = strrchr(args[0],'/')+1;
						execv(argv[optind+1],args);
					}
					else
						execl(SHELL,strrchr(SHELL,'/')+1,"-c",argv[optind+1],(char *) 0);

					/* normally, we will not get to this point */

					/* Get these strings into the binary */
					if (sccsid == copyright)
						(void)fprintf(stderr,"%s\n",sccsid);

					fprintf(stderr,"can't execute shell\n");
					exit(1);
		default	:	/* parent process: wait for the death of the child :-( */
					wait(&stat_val);
					exit(WIFSIGNALED(stat_val)?WTERMSIG(stat_val)+128:WEXITSTATUS(stat_val));
	}
}
